//************************************************************
// By Higor
// This channel was made to prevent unwanted function calls
// Keeps the log cleaner
//************************************************************
class XC_CompensatorChannel expands Info
	config(LCWeapons);

var XC_LagCompensation LCActor;
var XC_LagCompensator LCComp;
var PlayerPawn LocalPlayer;
var float ffRefireTimer; //This will enforce security checks
var float cAdv;
var float pwAdjust;
var float pwChain;
var Weapon CurWeapon, PendingWeapon;
var rotator OldView;
var bool bUseLC;
//If LC is globally disabled, this actor won't exist
var bool bDelayedFire;
var bool bLogTick;
var bool bJustSwitched;
var bool bFakeSwitch;
var bool bAlreadyProcessed; //Queue following shots

var() config bool bUseLagCompensation;

// Shoot flags
// 1 - Use PlayerCalcView
// 2 - Use ZP origin and dir instead of real origin and dir
// 4 - Add range (+10k)
// 8 - Add range (+20k)
// Last 16 bytes = random seed for aim error


struct ShotData
{
	var private Actor ffOther;
	var private Weapon Weap;
	var private int ffID, CmpRot, ShootFlags;
	var private float ffTime;
	var private vector ffHit, ffOff, ffStartTrace;
	var private rotator ffView;
	var private float ffAccuracy;
};
var private ShotData SavedShots[8];
var private int ffISaved;

replication
{
	reliable if ( Role == ROLE_Authority )
		bUseLC, cAdv;
	reliable if ( Role == ROLE_Authority )
		ffForceLC, ClientChangeLC, SetPendingW;
	reliable if ( Role < ROLE_Authority )
		ffSetLC, ffSendHit;
}

//Saves hit info, processes after movement physics occurs, do simplest of checks
function ffSendHit( Actor ffOther, Weapon Weap, int ffID, float ffTime, vector ffHit, vector ffOff, rotator ffView, vector ffStartTrace, int CmpRot, int ShootFlags, optional float ffAccuracy)
{
	if ( ffISaved > 8 || (Weap == none) || (CurWeapon != Weap) )
		return;
	if ( ffOther != none && ffID != -1 ) //Process this here
		return;
	if ( Weap.IsInState('DownWeapon') )
	{
		Log("DOWNWEAPON State", 'LagCompensator');
		return;
	}
	if ( int(ffAccuracy != 0) + int((ShootFlags >>> 16) == 0) != 1 )
	{
		Log("Accuracy("$ffAccuracy$") and ShootFlags("$ShootFlags$") inconsistency",'LagCompensator');
		return;
	}

	if ( !bAlreadyProcessed )
	{
		bAlreadyProcessed = true;
		if ( !ProcessHit( ffOther, Weap, ffID, ffTime, ffHit, ffOff, ffView, ffStartTrace, CmpRot, ShootFlags, ffAccuracy) )
			Goto TO_QUEUE;
		return;
	}
	TO_QUEUE: //Give this shot another opportunity to process at the end of the frame
	Log("Shot delayed to end of frame",'LagCompensator');
	SavedShots[ffISaved].ffOther = ffOther;
	SavedShots[ffISaved].Weap = Weap;
	SavedShots[ffISaved].ffID = ffID;
	SavedShots[ffISaved].ffTime = ffTime;
	SavedShots[ffISaved].ffHit = ffHit;
	SavedShots[ffISaved].ffOff = ffOff;
	SavedShots[ffISaved].ffView = ffView;
	SavedShots[ffISaved].ffStartTrace = ffStartTrace;
	SavedShots[ffISaved].CmpRot = CmpRot;
	SavedShots[ffISaved].ShootFlags = ShootFlags;
	SavedShots[ffISaved].ffAccuracy = ffAccuracy;
	ffISaved++;
}

function bool ProcessHit( Actor ffOther, Weapon Weap, int ffID, float ffTime, vector ffHit, vector ffOff, rotator ffView, vector ffStartTrace, int CmpRot, int ShootFlags, float ffAccuracy)
{
	local XC_LagCompensator ffLC;
	local vector X, Y, Z, EndTrace;
	local float Range;
	local byte Imprecise;
	local int Seed;

	Imprecise = byte(LCComp.ImpreciseTimer > 0);
	if ( ffAccuracy != 0 )
		Seed = ShootFlags >>> 16;
	if ( !bUseLC || !LCActor.FastValidate( self, ffOther, ffID, Weap, CmpRot, Owner.Location, ffStartTrace, ffHit, Imprecise, ffAccuracy, ShootFlags) ) //NEEDS FIXING!
		return false;
	if ( !LCComp.ffClassifyShot(ffTime) ) //Shot denied
	{	Log("Time classification failure");		return false;	}

	LCComp.ffCurRegTimer = float(Weap.GetPropertyText("ffRefireTimer")); //Register this timer as current shoot timer
	if ( (LCComp.ImpreciseTimer <= 0) && (Imprecise > 0) ) //Player skipped a security check
		LCComp.ImpreciseTimer = LCComp.ffCurRegTimer * 4;
	LCComp.ffRefireTimer += LCComp.ffCurRegTimer; //Register this new shot in the refire protection

	if  ( ffID != -1 )
		ffOther = LCComp.ffCheckHit( ffID, ffHit, ffOff, ffView);
	
	if ( (ffOther == none) || !Class'LCStatics'.static.RelevantHitActor(ffOther, PlayerPawn(Owner), float(LCComp.ffLastPing) / 1000.0) ) //Shot missed, get another target
	{
		if ( (ShootFlags & 2) == 0 )
		{
			GetAxes( Pawn(Owner).ViewRotation, X, Y, Z);
			if ( (ShootFlags & 1) == 0 )
				ffStartTrace = Owner.Location + Pawn(Owner).BaseEyeheight * vect(0,0,1);
			else
				ffStartTrace = Owner.Location + Weap.CalcDrawOffset() + Weap.FireOffset.Y * Y + Weap.FireOffset.Z * Z;
		}
		else
			GetAxes( rotator(ffHit - ffStartTrace), X, Y, Z);
		Range = 10000;
		if ( (ShootFlags & 4) > 0 )			Range += 10000;
		if ( (ShootFlags & 8) > 0 )			Range += 20000;
		EndTrace = ffStartTrace + X * Range;
		if ( Seed > 0 )
			EndTrace += class'LCStatics'.static.StaticAimError( Y, Z, ffAccuracy, Seed);
		ffOther = Class'LCStatics'.static.ffIrrelevantShot( ffHit, ffOff, EndTrace, ffStartTrace, Pawn(Owner), float(LCComp.ffLastPing) / 1000.0 );
		Weap.ProcessTraceHit( ffOther, ffHit, ffOff, X, Y, Z);
		return true;
	}
	ffHit = ffOther.Location + ffOff;
	GetAxes( rotator(ffHit - ffStartTrace), X, Y, Z);
	Weap.ProcessTraceHit( ffOther, ffHit, -X, X, Y, Z);
	return true;
}

function ffSetLC( bool bEnable)
{
	if ( bUseLC == bEnable )
		ffForceLC( bEnable);
	bUseLC = bEnable;
}

simulated event PostNetBeginPlay()
{
	if ( PlayerPawn(Owner) != none && ViewPort(PlayerPawn(Owner).Player) != none ) 
		LocalPlayer = PlayerPawn(Owner);
	else
		return;

//	if ( LocalPlayer.Weapon != none )    //Register existing weapon
	GotoState('ClientOp');
}

simulated state ClientOp
{
	simulated event Tick( float DeltaTime)
	{
		if ( LocalPlayer.Weapon != CurWeapon )
		{
			CurWeapon = LocalPlayer.Weapon;
			if ( CurWeapon != none )
			{
				CurWeapon.KillCredit( self);
				if ( CurWeapon.IsAnimating() )
				{
					if ( !bFakeSwitch )
						CurWeapon.AnimFrame = fMin( CurWeapon.AnimFrame + cAdv, 0.99);
					else if ( pwChain > 0 )
						CurWeapon.AnimFrame = fMin( CurWeapon.AnimFrame + pwChain, 0.99);
					pwChain = 0;
				}
			}
			bJustSwitched = (TournamentWeapon(CurWeapon) != none);
		}
		if ( bLogTick )
		{
			Log("Channel tick at "$Level.TimeSeconds);
			bLogTick = false;
		}
		if ( bDelayedFire )
			CurWeapon.KillCredit( self);
		if ( bJustSwitched && TournamentWeapon(CurWeapon).bCanClientFire )
		{
			bJustSwitched = false;
			if ( LocalPlayer.bFire > 0 )
				CurWeapon.ClientFire(0);
			else if ( LocalPlayer.bAltFire > 0 )
				CurWeapon.ClientAltFire(0);
			else
				bJustSwitched = true;
		}
		if ( pwAdjust > 0 ) //Pending weapon mechanics
			ClientPendingAdjust( DeltaTime);
	}
Begin:
	Spawn(class'LCBindScanner');
	if ( LocalPlayer.IsA('bbPlayer') )
		Spawn(class'LCWeaponHUD').LocalPlayer = LocalPlayer;
FindClient:
	Sleep(0.0);
AdjustClient:
	if ( bUseLagCompensation != bUseLC ) //This will work on high packet loss environments
		ffSetLC( bUseLagCompensation);
	Sleep(0.5);
	Goto('AdjustClient');
}

function AddPlayer( PlayerPawn Other, XC_LagCompensation Master)
{
	if ( (Other != none) && (NetConnection(Other.Player) != none) && (Master != none) )
	{
		SetOwner( Other);
		LCActor = Master;
		LCComp = LCActor.ffGetLC( Other);
		GotoState('ServerOp');
	}
	else
		Destroy();
}

state ServerOp
{
	event Tick( float DeltaTime)
	{
		local int i;
		While ( i < ffISaved )
		{
			CurWeapon.Tick(0);
			ProcessHit( SavedShots[i].ffOther, SavedShots[i].Weap, SavedShots[i].ffID, SavedShots[i].ffTime, SavedShots[i].ffHit, SavedShots[i].ffOff, SavedShots[i].ffView, SavedShots[i].ffStartTrace, SavedShots[i].CmpRot, SavedShots[i].ShootFlags, SavedShots[i].ffAccuracy);
			SavedShots[i].ffOther = none;
			SavedShots[i].ffID = -1;
			SavedShots[i].Weap = none;
			i++;
		}
		ffISaved = 0;
		bAlreadyProcessed = false;
		if ( LCActor.bWeaponAnim )
			cAdv = (float(LCComp.ffLastPing) / 1000) * Level.TimeDilation;
		if ( Owner != none )
			OldView = Pawn(Owner).ViewRotation;
	}
Begin:
	While ( Owner != none && !Owner.bDeleteMe )
	{
		if ( Pawn(Owner).Weapon != CurWeapon )
		{
			if ( CurWeapon != none )
				CurWeapon.SetPropertyText("LCChan","None");
			CurWeapon = Pawn(Owner).Weapon;
			if ( CurWeapon != none )
				CurWeapon.KillCredit( self);
		}
		if ( LCActor.bPendingWeapon )
		{
			if ( (Pawn(Owner).PendingWeapon != none) && (PendingWeapon == none || PendingWeapon != Pawn(Owner).PendingWeapon) )
			{
				PendingWeapon = Pawn(Owner).PendingWeapon;
				SetPendingW( PendingWeapon);
			}
		}
		Sleep(0.0);
	}
	Destroy();
}

event SetInitialState()
{
	bScriptInitialized = true;
}

event Destroyed()
{
	if ( (CurWeapon != none) && !CurWeapon.bDeleteMe )
		CurWeapon.SetPropertyText("LCChan","None");
}

simulated function ffForceLC( bool bEnable)
{
	bUseLC = bEnable;
}

simulated function ClientChangeLC( bool bEnable)
{
	bUseLagCompensation = bEnable;
	SaveConfig();
}

simulated function SetPendingW( weapon Other)
{
	if ( Other == none || Other.bDeleteMe || CurWeapon == Other || CurWeapon == none )
		return;
	LocalPlayer.PendingWeapon = Other;
	PendingWeapon = Other;
	pwAdjust = 2 + cAdv;
	if ( CurWeapon.IsInState('ClientDown') )
	{
		if ( CurWeapon.AnimFrame + cAdv >= 1 )
		{
			pwChain = CurWeapon.AnimFrame + cAdv - 1;
			CurWeapon.AnimFrame = 0.99;
		}
		else
			CurWeapon.AnimFrame = fMin( CurWeapon.AnimFrame + cAdv, 0.99);
		bFakeSwitch = true;
	}
}

simulated function ClientPendingAdjust( float DeltaTime)
{
	if ( PendingWeapon == LocalPlayer.Weapon )
		LocalPlayer.PendingWeapon = none;
	if ( (pwAdjust -= DeltaTime) <= 0 ) //Reset
	{
		bFakeSwitch = false;
		pwAdjust = 0;
		PendingWeapon = none;
		LocalPlayer.PendingWeapon = none;
	}
}


defaultproperties
{
    bGameRelevant=True
    bHidden=True
    NetPriority=1.1
    NetUpdateFrequency=20
    RemoteRole=ROLE_SimulatedProxy
    bUseLC=True
    bUseLagCompensation=True
}